import argparse, textwrap
import sys
import os
import re
import logging
import configparser
import time
import stat
from subprocess import check_output, CalledProcessError

from .ubxneom8 import UbxNeoM8Receiver, UbxSetupFailed, ConfigurationException
from .logger import log

MIN_ADR_COMPATIBLE_VERSION = 4.21
MIN_ADR_RECOMMENDED_VERSION = 4.31
MAX_ADR_COMPATIBLE_VERSION = MIN_ADR_RECOMMENDED_VERSION

CRT_CFG_VERSION = 5

class DeviceInUseException(Exception):
    """ Raised if the device is already used by another process """

class NotExistingDevice(Exception):
    """ Raised when trying to initialize a non existing device """

class MissingConfigFile(Exception):
    """ Raised when the config file does not exist """

class GnssMgr():
    def __init__(self, name, configure_bitrate=False):
        device_path = f'/dev/{name}'

        # Check that device exists
        if not os.path.exists(device_path):
            log.error("Device does not exist")
            raise NotExistingDevice()
        if not stat.S_ISCHR(os.stat(device_path)[stat.ST_MODE]):
            log.error("Device is not a character device")
            raise NotExistingDevice()

        # Check that device is free from other tools
        try:
            pid = check_output(["fuser" , device_path]).rstrip()
            log.error("Another process (PID: %s) is accessing the receiver.", pid.decode())
            raise DeviceInUseException()
        except CalledProcessError:
            # fuser is returning 0 where this is actually a file user
            # and an error when there is no
            pass

        self.name = name
        self.receiver = None
        self.config = configparser.ConfigParser()
        self.setup_receiver(configure_bitrate)

    def stop(self):
        self.receiver.cleanup()
        self.receiver = None

    def setup_receiver(self, configure_bitrate=False):
        log.info("Connecting to receiver")
        max_con_retrials = 5
        retrials = 0
        self.receiver = None
        while retrials < max_con_retrials:
            retrials += 1
            try:
                self.receiver = UbxNeoM8Receiver(f'/dev/{self.name}', 115200, configure_bitrate)
                break
            except UbxSetupFailed as e:
                log.error("Initializing receiver failed: %s", str(e))
                time.sleep(1)

        if self.receiver is None:
            log.error("Could not establish a link with the GNSS receiver")
            sys.exit(1)
        log.debug("Needed " + str(retrials) + " retrial(s) for connecting to GNSS receiver")

    # ---------------------------------------------------------------------------------------
    def set_nmea_version(self, version):
        self.receiver.set_nmea_version(version)

        # Changing NMEA version may make the receiver unresponsive for some time
        # Trying to read back until it works again
        while True:
            try:
                time.sleep(0.5)
                self.receiver.get_nmea_version()
                break
            except:
                pass

    # ---------------------------------------------------------------------------------------
    def set_ini_time(self):
        self.receiver.set_receiver_init_time()

    # ---------------------------------------------------------------------------------------
    def reset_receiver(self, isRestart, isClear):
        if isClear:
            log.info("Reset of receiver's configuration")
            self.receiver.reset_config()
            log.info("Configuration reset, device restarting...")
            log.info("Reconnecting to receiver")
            self.stop()
            self.setup_receiver(True)
        elif isRestart:
            log.info("Cold booting...")
            self.receiver.coldboot_receiver()
            log.info("Cold booting ongoing...")
            log.info("Reconnecting to receiver")
            self.stop()
            self.setup_receiver(True)


    # ---------------------------------------------------------------------------------------
    def save_state(self, isPersist, isVerify, isClear):
        if isPersist:
            log.info("Persisting receiver state")
            #self.receiver.stop_gnss_function()
            self.receiver.save_receiver_state()
        if isVerify:
            log.info("Verifying restore of receiver state")
            self.receiver.verify_restoring_state()
        if isClear:
            log.info("Clear saved receiver state")
            self.receiver.clear_receiver_state()



    # ---------------------------------------------------------------------------------------
    def get_config(self, section, fieldname, experimental=False):
        try:
            value = self.config.get(section, fieldname) or None
        except configparser.NoOptionError:
            value = None
            if experimental:
                log.debug(f"[{section}] {fieldname} not set in config file")
            else:
                log.warning(f"[{section}] {fieldname} not set in config file")
        return value


    # ---------------------------------------------------------------------------------------
    def set_config_default(self):
        update_rate = self.get_config('default', 'update-rate')

        if update_rate:
            update_rate = int(update_rate)
            self.receiver.set_update_rate(update_rate)
            log.info("Update rate successfully configured")
        else:
            log.info("Update rate not configured")


    def read_config_default(self):
        update_rate = self.receiver.get_update_rate()
        self.config['default'].update({'update-rate': str(update_rate)})

    # ---------------------------------------------------------------------------------------
    def read_config_alignment(self):
        yaw, pitch, roll = self.receiver.get_alignment()
        self.config['installation'].update({'yaw': str(yaw),
                                       'pitch': str(pitch),
                                       'roll': str(roll)})

    def set_config_alignment(self):

        yaw = self.get_config('installation', 'yaw')
        pitch = self.get_config('installation', 'pitch')
        roll = self.get_config('installation', 'roll')

        if yaw:
            try:
                yaw = int(yaw)
            except ValueError:
                log.error("Invalid value for yaw: %s, ignoring", yaw)
                yaw = None
        if pitch:
            try:
                pitch = int(pitch)
            except ValueError:
                log.error("Invalid value for pitch: %s, ignoring", pitch)
                pitch = None
        if roll:
            try:
                roll = int(roll)
            except ValueError:
                log.error("Invalid value for roll: %s, ignoring", roll)
                roll = None

        if yaw is not None or pitch is not None or roll is not None:
            self.receiver.set_alignment(yaw, pitch, roll)
            log.info("yaw/pitch/roll successfully configured")
        else:
            log.info("yaw/pitch/roll not configured")

    # ---------------------------------------------------------------------------------------
    def read_config_positioning(self):
        vrp2antenna, vrp2imu = self.receiver.get_positioning()
        vrp2antenna = ';'.join([str(x) for x in vrp2antenna])
        vrp2imu = ';'.join([str(x) for x in vrp2imu])

        self.config['installation'].update({'vrp2antenna': vrp2antenna,
                                            'vrp2imu': vrp2imu})

    def set_config_positioning(self):
        vrp2antenna = self.get_config('installation', 'vrp2antenna')
        vrp2imu = self.get_config('installation', 'vrp2imu')

        if vrp2antenna:
            try:
                vrp2antenna = [float(x) for x in vrp2antenna.split(';')]
                if len(vrp2antenna) != 3:
                    raise ValueError("3 values required : x,y,z")
            except ValueError:
                log.error("vrp2antenna is not valid: %s, expect format: x;y;z. Dot-decimal notation for x, y and z", vrp2antenna)
                vrp2antenna = None

        if vrp2imu:
            try:
                vrp2imu = [float(x) for x in vrp2imu.split(';')]
                if len(vrp2imu) != 3:
                    raise ValueError("3 values required : x,y,z")
            except ValueError:
                log.error("vrp2imu is not valid: %s, expect format: x;y;z. Dot-decimal notation for x, y and z", vrp2antenna)
                vrp2imu = None

        if vrp2antenna or vrp2imu:
            self.receiver.set_positioning(vrp2antenna, vrp2imu)
            log.info("Lever arms successfully configured")
        else:
            log.info("Lever arms not configured")

    # ---------------------------------------------------------------------------------------
    def read_config_navigation(self):
        dyn_model, fix_mode = self.receiver.get_nav_engine()
        high_navigation_rate = self.receiver.get_high_navigation_rate()
        self.config['navigation'] = {'mode': dyn_model,
                                     'fix-mode': fix_mode,
                                     'high-update-rate': high_navigation_rate
                                    }


    def set_config_navigation(self):
        dead_reckoning = self.get_config('navigation', 'dead-reckoning')

        if dead_reckoning is not None:
            self.receiver.set_navx5(True if dead_reckoning == "enable" else False)

        high_navigation_rate = self.get_config('navigation', 'high-update-rate')
        if high_navigation_rate:
            high_navigation_rate = int(high_navigation_rate)
            self.receiver.set_high_navigation_rate(high_navigation_rate)

        dyn_model = self.get_config('navigation', 'mode')
        fix_mode = self.get_config('navigation', 'fix-mode', experimental=True)

        if dyn_model or fix_mode:
            self.receiver.set_nav_engine(dyn_model, fix_mode)

        gnss_systems = self.get_config('navigation', 'systems')

        if gnss_systems:
            gps = False
            glonass = False
            beidou = False
            galileo = False
            sbas = False
            imes = False
            qzss = False

            for system in gnss_systems.lower().split(';'):
                if system == 'gps':
                    gps = True
                elif system == 'glonass':
                    glonass = True
                elif system == 'beidou':
                    beidou = True
                elif system == 'galileo':
                    galileo = True
                elif system == 'sbas':
                    sbas = True
                elif system == 'imes':
                    imes = True
                elif system == 'qzss':
                    qzss = True
                else:
                    log.error("Unsupported gnss system: %s, ignoring", system)
            try:
                self.receiver.set_gnss_systems(gps, glonass, galileo, beidou, sbas, imes, qzss)
                log.info("GNSS systems successfully configured")
            except ConfigurationException:
                log.error("Unsupported navigation systems combination, ignoring: %s", ';'.join(gnss_systems))


        else:
            log.info("GNSS systems not configured")

    # ---------------------------------------------------------------------------------------

    def set_config_time(self):
        enable = self.get_config('time', 'timepulse')
        if enable is not None:
            if enable == "enable":
                enable = True
            elif enable == "disable":
                enable = False
            else:
                log.error(f"Unsupported value for [time] timepulse: {enable}, not configuring")
                enable = None

        freq = self.get_config('time', 'timepulse-frequency')
        if freq is not None:
            try:
                freq = int(freq)
                # The receiver is configured in us, which means that the
                # shortest low and high durations are both 1us
                # which make the shortest period 2us
                # And therefore a max frequency of 500kHz
                if freq < 1 or freq > 500000:
                    log.error(f"Value out of range for [time] timepulse-frequency: {freq}, not configuring")
                    freq = None
            except ValueError:
                log.error(f"Non numeric value for [time] timepulse-frequency: {freq}, not configuring")
                freq = None


        if enable is not None or freq is not None:
            self.receiver.set_timepulse(enable, freq)
        else:
            log.info("Time configuration not configured")

    def read_config_time(self):
        enable, freq = self.receiver.get_timepulse()
        self.config['time'] = {'timepulse': 'enable' if enable else "disable",
                               'timepulse-frequency': freq }



    # ---------------------------------------------------------------------------------------

    def check_config_version(self):
        try:
            version = self.config.get('default', 'version')
            version = int(version)
        except (configparser.NoOptionError, configparser.NoSectionError):
            log.error("No config version, assuming version 1")
            version = 1

        if version == CRT_CFG_VERSION:
            log.info("Config file version is correct")
            return

        if version < CRT_CFG_VERSION:
            log.warning("The config file version is obsolete, upgrading is recommended")
        elif version > CRT_CFG_VERSION:
            log.warning("The config file version is more recent than the supported one. Correct behavior cannot be guaranteed")
        log.warning("Consider using the config file template in /usr/etc/gnss/%s.conf", self.name)


    # ---------------------------------------------------------------------------------------
    def configure_receiver(self):
        if 'default' in self.config:
            self.set_config_default()
        if 'installation' in self.config:
            self.set_config_alignment()
            self.set_config_positioning()
        if 'navigation' in self.config:
            self.set_config_navigation()
        if 'time' in self.config:
            self.set_config_time()

    def read_config(self):
        self.config['default'] = { 'version': str(CRT_CFG_VERSION) }
        self.config['installation'] = {}
        self.read_config_default()
        self.read_config_alignment()
        self.read_config_positioning()
        self.read_config_navigation()
        self.read_config_time()

    # ---------------------------------------------------------------------------------------
    def configuration(self, cfgFile=None, destFile=""):
        if cfgFile is None:
            cfgFile = f'/etc/gnss/{self.name}.conf'

        if not os.path.exists(cfgFile):
            log.error("Config file does not exist")
            raise MissingConfigFile()

        log.info("Reading configuration from %s", cfgFile)

        doCfgRead = (destFile != "")
        if doCfgRead:
            self.read_config()
            with open(destFile, "w") as fd:
                self.config.write(fd)
            log.info("Configuration written to " + destFile)
        else:
            try:
                self.config.read(cfgFile)
            except configparser.DuplicateOptionError as e:
                log.error("Duplicated option: %s", str(e))
                self.config = configparser.ConfigParser(strict=False)
                self.config.read(cfgFile)

            self.check_config_version()
            self.configure_receiver()

    def persist_configuration(self):
        self.receiver.persist_receiver_config()


    # ---------------------------------------------------------------------------------------
    def get_and_write_info(self, infofile=None):
        if infofile is None:
            infofile = f'/run/gnss/{self.name}.config'

        log.info("Writing info from receiver to %s", infofile)

        info = self.receiver.get_info()
        # TODO: move to ubxneom8.py since it is highly receiver specific
        # TODO: what about non ADR fw types?
        global MIN_ADR_RECOMMENDED_VERSION
        global MIN_ADR_COMPATIBLE_VERSION
        global MAX_ADR_COMPATIBLE_VERSION

        active_dead_reckoning_abbreviation = 'ADR'
        standard_precision_gnss_abbreviation = 'SPG'

        firmware_append = ''
        regex_pattern_fw_info = re.compile('(?P<fw_type>\w*) (?P<fw_version>[\d.]*)')
        fw_info = regex_pattern_fw_info.search(info['fw'])
        # TODO: get rid of nested if
        if fw_info:
            # check for ADR fw type
            if fw_info.group('fw_type') == active_dead_reckoning_abbreviation:

                if float(fw_info.group('fw_version')) < MIN_ADR_COMPATIBLE_VERSION:
                    firmware_append = ' (Obsolete)'
                elif float(fw_info.group('fw_version')) < MIN_ADR_RECOMMENDED_VERSION:
                    firmware_append = ' (Deprecated)'
                elif float(fw_info.group('fw_version')) > MAX_ADR_COMPATIBLE_VERSION:
                    firmware_append = ' (Untested)'
            elif fw_info.group('fw_type') == standard_precision_gnss_abbreviation:
                firmware_append = ' (Deprecated)'
            else:
                log.info("unknown fw type detected")
                firmware_append = ' (Untested)'
        else:
            log.warning("unable to check if fw version is up to date")
            firmware_append = ' (Untested)'

        if os.path.dirname(infofile) != '':
            os.makedirs(os.path.dirname(infofile), exist_ok=True)
        fd = open(infofile, "w")
        fd.write(
            f"{'Vendor:':<36}{info['vendor']}\n"
            f"{'Model:':<36}{info['model']}\n"
            f"{'Firmware:':<36}{info['fw']}{firmware_append}\n"
            f"{'ubx-Protocol:':<36}{info['ubxprot']}\n"
            f"{'Supported Satellite Systems:':<36}{info['sat']}\n"
            f"{'Supported Augmentation Services:':<36}{info['svc']}\n"
            f"{'SW Version:':<36}{info['sw']}\n"
            f"{'HW Version:':<36}{info['hw']}\n"
        )
        fd.close()


    # ---------------------------------------------------------------------------------------
    def perform_startup_steps(self, args):
        log.info("Perform startup steps: setup com and receiver; get and write info; check receiver state (SoS)")
        self.configuration(args.file)
        self.get_and_write_info(args.info)
        self.save_state(False, True, True)

