#!/usr/bin/env python3
"""
gnss-mgr CLI parser skeleton
"""
import sys
import argparse
import logging

from .gnss_mgr import GnssMgr, DeviceInUseException, NotExistingDevice, MissingConfigFile
from .platform import Platform

from .__version__ import __version__
from .logger import log


class GnssMgrCLI:
    def __init__(self):
        self.parser = None

    def setup_parser(self):
        # ============================================================================
        # setup the argument parser
        # - general part
        # ============================================================================

        parser = argparse.ArgumentParser(
            description='Manages GNSS modem')

        parser.add_argument(
            'device',
            help='local serial device to which GNSS modem is connected (e.g. /dev/gnss0)')

        parser.add_argument(
            '-V', '--version',
            action='version',
            version='%(prog)s (v' + __version__ + ')')

        parser.add_argument(
            '-v', '--verbose',
            action='store_true',
            dest='verbose',
            help='be verbose, show debug output')

        parser.add_argument(
            '-q', '--quiet',
            action='store_true',
            dest='quiet',
            help='be quiet, only show warnings and errors')

        # ============================================================================
        # register gnss-mgr sub commands
        # ============================================================================

        subparser = parser.add_subparsers(
            title='command',
            help='select command',
            dest='command')
        subparser.required = True

        self.setup_parser_init(subparser)
        self.setup_parser_sos(subparser)
        self.setup_parser_config(subparser)
        self.setup_parser_control(subparser)

        # Remember parser for later usage
        self.parser = parser

    def setup_parser_init(self, parser_top):
        sub_parser = parser_top.add_parser(
            'init',
            help='sets up GNSS modem')  # TODO

        sub_parser.add_argument(
            '-f', '--file',
            help='path to run file',
            dest='runfile')

    def setup_parser_sos(self, parser_top):
        sub_parser = parser_top.add_parser(
            'sos',
            help='save on shutdown operations')

        sub_parser.add_argument(
            'operation',
            choices=['save', 'clear'],
            help="selects sos operation to perform")

    def setup_parser_config(self, parser_top):
        sub_parser = parser_top.add_parser(
            'config',
            help='configures GNSS modem')

        sub_parser.add_argument(
            '-f', '--file',
            help='path to config  file',
            dest='configfile')

    def setup_parser_control(self, parser_top):
        sub_parser = parser_top.add_parser(
            'control',
            help='performs GNSS modem control function')

        sub_parser.add_argument(
            'action',
            choices=['cold-start', 'persist', 'factory-reset'],
            help="selects action to perform")


def build_name_from_device(device_name):
    """
    Takes device name of the form /dev/gnss0 and converts it to gnss0
    """
    device_prefix = '/dev/'

    if device_name.startswith(device_prefix):
        return device_name[len(device_prefix):]
    else:
        return device_name
    return


def main():
    rc = 0

    # create provision object
    gmcli = GnssMgrCLI()

    # parse arguments and provide necessary variables
    gmcli.setup_parser()

    args = gmcli.parser.parse_args()

    # setup logger based on -v, -q option
    if args.verbose:
        log.setLevel(logging.INFO)
    elif args.quiet:
        log.setLevel(logging.ERROR)
    else:
        log.setLevel(logging.WARNING)

    device = build_name_from_device(args.device)

    try:
        if 'init' in args.command:
            gnss = GnssMgr(device, configure_bitrate=True)
        else:
            gnss = GnssMgr(device, False)
    except (DeviceInUseException, NotExistingDevice):
        log.error("Aborting")
        sys.exit(-1)

    # execute command
    # TODO: Factor out functions to dedicated modules/classes
    # Implement relevant handle function in these modules/classes
    if 'init' in args.command:
        gnss.get_and_write_info(args.runfile)
        if Platform.is_sos_enabled():
            gnss.save_state(False, True, True)
            # Time assistance is required for dead reckoning when no satelite is visible
            gnss.set_ini_time()

        # NMEA version 4.1 is required to use Galileo positioning
        gnss.set_nmea_version(4.1)

    elif 'sos' in args.command:
        if 'save' in args.operation:
            gnss.save_state(True, False, False)
        elif 'clear' in args.operation:
            gnss.save_state(False, False, True)

    elif 'config' in args.command:
        try:
            gnss.configuration(args.configfile)
            log.info("Configuration successful")
        except MissingConfigFile:
            sys.exit(-1)

    elif 'control' in args.command:
        if 'cold-start' in args.action:
            gnss.reset_receiver(True, False)
        elif 'persist' in args.action:
            gnss.persist_configuration()
        elif 'factory-reset' in args.action:
            gnss.reset_receiver(False, True)

    else:
        # due to argparser checks we should never get here
        rc = 10

# To test parser programmatically
#    gm.parser.parse_args(['/dev/gnss0', 'init'])

    gnss.stop()
    sys.exit(rc)


if __name__ == '__main__':
    main()
