
from nmubxlib.ubx_cfg_esfalg import UbxCfgEsfAlg  # pylint: disable=E0401
from nmubxlib.ubx_cfg_esfgawt import UbxCfgEsfGawt  # pylint: disable=E0401

from .logger import log


class ArmLeverNotFound(Exception):
    """ Raised if a configured arm lever is not found in the config of the receiver """

class InvalidValue(Exception):
    """ Raised when config file does not contain a valid value """

class UbxUtils():

    @staticmethod
    def find_arm_by_name(full_cfg, arm_name):
        arm_id = None

        # Convert arm name to arm number
        # Order is important: each arm has a matching id: 0 for vrp2antenna...
        arm_names = ['vrp2antenna', 'vrp2imu', 'imu2antenna', 'imu2vrp', 'imu2crp']
        for i, name in enumerate(arm_names):
            if arm_name == name:
                arm_id = i
                break

        if arm_id is None:
            raise ArmLeverNotFound()

        # Find position of arm with this id
        pos = 0
        while True:
            try:
                arm_id_cfg = full_cfg.f.get(f'leverArmType_{pos}').value
                if arm_id_cfg == arm_id:
                    return pos
                pos += 1
            except:
                raise ArmLeverNotFound()


    @staticmethod
    def range_check(value, min, max):
        if value > max or value < min:
            raise InvalidValue()


    @staticmethod
    def get_wheeltick_flags_string(flags):
        wtFlags = dict()
        wtFlags["useWtPin"] = 0
        if (flags & UbxCfgEsfGawt.MASK_useWtPin) != 0:
            wtFlags["useWtPin"] = 1
        # wtFlags["useWtSpeed"] = 0
        # if (flags & UbxCfgEsfGawt.MASK_useWtSpeed) != 0:
        #    wtFlags["useWtSpeed"] = 1
        return wtFlags


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def adapt_wheeltick_flags(actualFlags, flags):
        for flagItem in flags:
            if flagItem == "useWtPin":
                if flags[flagItem] == 0:
                    actualFlags &= ~UbxCfgEsfGawt.MASK_useWtPin
                else:
                    actualFlags |= UbxCfgEsfGawt.MASK_useWtPin
            elif flagItem == "useWtSpeed":
                if flags[flagItem] == 0:
                    actualFlags &= ~UbxCfgEsfGawt.MASK_useWtSpeed
                else:
                    actualFlags |= UbxCfgEsfGawt.MASK_useWtSpeed
            else:
                log.warning("Unsupported wheel tick flag item: " + str(flagItem))
        return actualFlags


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def get_alignment_bitfield_object(bitfield):
        bitfieldCfg = dict()
        bitfieldCfg["doAutoMntAlg"] = 0
        if (bitfield & UbxCfgEsfAlg.BITFIELD_doAutoMntAlg) != 0:
            bitfieldCfg["doAutoMntAlg"] = 1
        return bitfieldCfg


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def adapt_alignment_bitfield(actualField, field):
        for fieldItem in field:
            if fieldItem == "doAutoMntAlg":
                if field[fieldItem] == 0:
                    actualField &= ~UbxCfgEsfAlg.BITFIELD_doAutoMntAlg
                else:
                    actualField |= UbxCfgEsfAlg.BITFIELD_doAutoMntAlg
            else:
                log.warning("Unsupported alignment bitfield item: " + str(fieldItem))
        return actualField


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def nav_engine_model_value2string(value):
        if value == 0:
            return "portable"
        if value == 2:
            return "stationary"
        if value == 3:
            return "pedestrian"
        if value == 4:
            return "vehicle"
        if value == 5:
            return "sea"
        if value == 6:
            return "airborne <1g"
        if value == 7:
            return "airborne <2g"
        if value == 8:
            return "airborne <4g"
        if value == 9:
            return "wrist worn watch"
        if value == 10:
            return "bike"
        raise InvalidValue("Unsupported nav engine model value: %d", value)


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def nav_engine_model_string2value(model_string):
        if model_string == "portable":
            return 0
        if model_string == "stationary":
            return 2
        if model_string == "pedestrian":
            return 3
        if model_string == "vehicle" or model_string == "automotive":
            return 4
        if model_string == "sea":
            return 5
        if model_string == "airborne <1g":
            return 6
        if model_string == "airborne <2g":
            return 7
        if model_string == "airborne <4g":
            return 8
        if model_string == "wrist worn watch":
            return 9
        if model_string == "bike":
            return 10
        raise InvalidValue("Unsupported nav engine model: %s", model_string)


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def get_nav_engine_fix_mode_string(value):
        if value == 1:
            return "2D"
        if value == 2:
            return "3D"
        if value == 3:
            return "auto"
        raise InvalidValue("Unsupported nav engine fix mode: " + value)


    # ---------------------------------------------------------------------------------------
    @staticmethod
    def get_nav_engine_fix_mode_value(mode_string):
        if mode_string == "2D":
            return 1
        if mode_string == "3D":
            return 2
        if mode_string == "auto":
            return 3
        raise InvalidValue("Unsupported nav engine fix mode: " + mode_string)
