""" Module to access and control ublox gnss receivers """
# pylint: disable=C0103
import time
import os
import datetime

from nmubxlib.server_tty import GnssUBlox # pylint: disable=E0401
from nmubxlib.ubx_ack import UbxAckAck # pylint: disable=E0401
from nmubxlib.ubx_upd_sos import UbxUpdSosPoll, UbxUpdSos, UbxUpdSosAction # pylint: disable=E0401
from nmubxlib.ubx_cfg_rst import UbxCfgRstAction # pylint: disable=E0401
from nmubxlib.ubx_cfg_nav5 import UbxCfgNav5Poll, UbxCfgNav5 # pylint: disable=E0401
from nmubxlib.ubx_cfg_navx5 import UbxCfgNavx5Poll, UbxCfgNavx5 # pylint: disable=E0401
from nmubxlib.ubx_cfg_nmea import UbxCfgNmeaPoll, UbxCfgNmea # pylint: disable=E0401
from nmubxlib.ubx_cfg_esfgawt import UbxCfgEsfGawt, UbxCfgEsfGawtPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_esfla import UbxCfgEsfla, UbxCfgEsflaPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_esfalg import UbxCfgEsfAlg, UbxCfgEsfAlgPoll # pylint: disable=E0401
from nmubxlib.ubx_mon_ver import UbxMonVer, UbxMonVerPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_prt import UbxCfgPrtUart, UbxCfgPrtPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_cfg import UbxCfgCfgAction # pylint: disable=E0401
from nmubxlib.ubx_cfg_rate import UbxCfgRate, UbxCfgRatePoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_hnr import UbxCfgHnr, UbxCfgHnrPoll # pylint: disable=E0401
from nmubxlib.ubx_cfg_gnss import UbxCfgGnss, UbxCfgGnssPoll # pylint: disable=E0401
from nmubxlib.ubx_mga_ini_time_utc import UbxMgaIniTimeUtc # pylint: disable=E0401
from nmubxlib.ubx_cfg_tp5 import UbxCfgTp5, UbxCfgTp5Poll  # pylint: disable=E0401

from .logger import log
from .ubx_utils import UbxUtils, InvalidValue


MICROSECONDS_PER_SECOND = 1000000

class PollingException(Exception):
    """ Raised  when polling data from the receiver fails """

class UbxSetupFailed(Exception):
    """ Raised when initial setup failed"""

class ConfigurationException(Exception):
    """ Raised when applying a configuration to the receiver failed """

class UnexpectedResponseException(Exception):
    """ Raised when the receiver answer with an unexpected answer """

class UbxNeoM8Receiver():
    """ Class to access and control ublox gnss receivers (NeoM8) """

    TP5_FLAG_ACTIVE = (1 << 0)
    TP5_FLAG_LOCK_GPS_FREQ = (1 << 1)
    TP5_FLAG_IS_LEN = (1 << 4)

    def __init__(self, tty, bitrate, configure_bitrate=True):
        self.tty = tty
        self.bitrate = bitrate
        self.ubx = None


        while True:
            try:
                # Setup ubx accessor
                self.__setup_ubx(bitrate, configure_bitrate)
                break
            except UbxSetupFailed:
                if not configure_bitrate:
                    return
                time.sleep(0.5)


    def __get_current_bitrate(self):
        log.info("Detecting bitrate")
        BITRATES_TO_TEST = [ 115200, 9600 ]
        while True:
            for bitrate in BITRATES_TO_TEST:
                self.ubx.set_baudrate(bitrate)
                if self.ubx.scan():
                    log.info("Bitrate detected: %d", bitrate)
                    return bitrate

            # Bitrate detection failed, waiting and retrying
            time.sleep(0.5)


    def __read_and_setup_bitrate(self, bitrate):
        crt_bitrate = self.__get_current_bitrate()
        while crt_bitrate != bitrate:
            # Configure requested bitrate
            self.config_com_bitrate(bitrate)
            log.warning("Receiver bitrate was %d, changed to %d", crt_bitrate, bitrate)
            # Check that bitrate has really changed
            crt_bitrate = self.__get_current_bitrate()


    def __setup_ubx(self, bitrate, configure_bitrate):
        self.ubx = GnssUBlox(self.tty, bitrate)
        ready = self.ubx.setup()
        if not ready:
            self.ubx.cleanup()
            del self.ubx
            self.ubx = None
            raise UbxSetupFailed("UBX setup failed")

        # Registering all used frames
        self.ubx.register_frame(UbxUpdSos)
        self.ubx.register_frame(UbxCfgRstAction)
        self.ubx.register_frame(UbxCfgCfgAction)
        self.ubx.register_frame(UbxMgaIniTimeUtc)
        self.ubx.register_frame(UbxCfgGnss)
        self.ubx.register_frame(UbxCfgRate)
        self.ubx.register_frame(UbxCfgHnr)
        self.ubx.register_frame(UbxCfgPrtUart)
        self.ubx.register_frame(UbxCfgNav5)
        self.ubx.register_frame(UbxCfgNavx5)
        self.ubx.register_frame(UbxCfgNmea)
        self.ubx.register_frame(UbxCfgEsfGawt)
        self.ubx.register_frame(UbxCfgEsfla)
        self.ubx.register_frame(UbxCfgEsfAlg)
        self.ubx.register_frame(UbxCfgTp5)
        self.ubx.register_frame(UbxMonVer)

        # Setup bitrate
        if configure_bitrate:
            self.__read_and_setup_bitrate(bitrate)

        # Try to poll data from receiver
        err_cnt = 0
        while err_cnt < 3:
            try:
                rawinfo = self._poll_information()
                log.info("Connected to GNSS receiver")
                return
            except PollingException:
                log.warning("Polling info failed, retrying")
                err_cnt += 1
                time.sleep(0.5)

        self.ubx.cleanup()
        self.ubx = None
        raise UbxSetupFailed("Poll not working")



    def cleanup(self):
        if self.ubx is None:
            return
        self.ubx.cleanup()
        log.debug("Communication to GNSS receiver cleaned up")
        del self.ubx
        self.ubx = None


    # ---------------------------------------------------------------------------------------
    #   Helper Functions
    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    def __poll(self, poller):
        err_cnt = 0
        while err_cnt < 3:
            cfg = self.ubx.poll(poller)
            if cfg is None:
                err_cnt += 1
                time.sleep(0.5)
            else:
                return cfg

        if cfg is None:
            raise PollingException("Polling failed for " + str(poller))


    def _poll_gnss_cfg(self):
        return self.__poll(UbxCfgGnssPoll())

    def _poll_update_rate(self):
        return self.__poll(UbxCfgRatePoll())

    def _poll_hnr(self):
        return self.__poll(UbxCfgHnrPoll())

    def _poll_port(self):
        port_poller = UbxCfgPrtPoll()
        port_poller.f.PortId = UbxCfgPrtPoll.PORTID_Uart
        return self.__poll(port_poller)


    # ---------------------------------------------------------------------------------------
    def _poll_save_on_shutdown(self):
        return self.__poll(UbxUpdSosPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_nav_engine(self):
        return self.__poll(UbxCfgNav5Poll())


    # ---------------------------------------------------------------------------------------
    def _poll_navx5(self):
        return self.__poll(UbxCfgNavx5Poll())


    # ---------------------------------------------------------------------------------------
    def _poll_nav_protocol(self):
        return self.__poll(UbxCfgNmeaPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_gyro_acc_and_wheeltick(self):
        return self.__poll(UbxCfgEsfGawtPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_positioning(self):
        return self.__poll(UbxCfgEsflaPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_alignment(self):
        return self.__poll(UbxCfgEsfAlgPoll())


    # ---------------------------------------------------------------------------------------
    def _poll_information(self):
        return self.__poll(UbxMonVerPoll())


    # ---------------------------------------------------------------------------------------
    def _sos_action(self, action):
        sos = UbxUpdSosAction()
        sos.f.cmd = action
        writeResp = self.ubx.set(sos)
        if writeResp is None:
            log.error("Not reponse from nmubxlib when saving or clearing receiver state: " + str(writeResp))
            raise ConfigurationException("Could not save or clear receiver state")

    def save_receiver_state(self):
        self._sos_action(UbxUpdSosAction.SAVE)
        log.info("Saving receiver state successfully performed")

    def clear_receiver_state(self):
        self._sos_action(UbxUpdSosAction.CLEAR)
        log.info("Clearing receiver state successfully performed")

    # ---------------------------------------------------------------------------------------
    def verify_restoring_state(self):
        sos_state = self._poll_save_on_shutdown()
        cmdField = sos_state.f.cmd
        response = sos_state.f.response
        if cmdField != 3:
            log.error("Restore verification got not the desired response: cmd=" + str(cmdField))
            raise UnexpectedResponseException("Unexpected response when verifying restore")
        state = "Unknown"
        if response == 1:
            state = "Failed"
            log.warning("Receiver restoring state: " + state)
        elif response == 2:
            state = "Success"
            log.info("Receiver restoring state: " + state)
        elif response == 3:
            state = "Nothing to restore"
            log.info("Receiver restoring state: " + state)
        else:
            log.warning("Receiver restoring state: " + state)


    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    #   Main NEOM8 Functions
    # ---------------------------------------------------------------------------------------
    def stop_gnss_function(self):
        rst = UbxCfgRstAction()
        rst.stop()
        rst.pack()
        self.ubx.fire_and_forget(rst)
        log.info("Stopping gnss function")

    def start_gnss_function(self):
        rst = UbxCfgRstAction()
        rst.start()
        rst.pack()
        self.ubx.fire_and_forget(rst)

    def coldboot_receiver(self):
        rst = UbxCfgRstAction()
        rst.cold_start()
        rst.pack()
        self.ubx.fire_and_forget(rst)
        log.info("Cold boot of GNSS receiver triggered, let receiver start")
        time.sleep(5)


    # ---------------------------------------------------------------------------------------
    def reset_config(self):
        cfgrst = UbxCfgCfgAction()
        cfgrst.reset(UbxCfgCfgAction.MASK_All)
        cfgrst.pack()
        self.ubx.fire_and_forget(cfgrst)
        log.info("Reset GNSS receiver configuration to default, let receiver start with default config")
        time.sleep(2)


    # ---------------------------------------------------------------------------------------
    def config_com_bitrate(self, bitrate):
        portcfg = self._poll_port()
        log.info("Setting communication bitrate from " + str(portcfg.f.baudRate) + " to " + str(bitrate))
        portcfg.f.baudRate = bitrate

        # Note: usually we would use ubx.set() to send the request and wait for an ack
        # Unfortunately the modem changes the bitrate immediately so the ACK is not
        # received.
        portcfg.pack()
        self.ubx.fire_and_forget(portcfg)
        log.info("Set default bitrate to " + str(bitrate) + " bit/s")

    # ---------------------------------------------------------------------------------------
    def set_receiver_init_time(self):
        mgainitime = UbxMgaIniTimeUtc()
        dt = datetime.datetime.utcnow()
        mgainitime.set_datetime(dt)
        self.ubx.fire_and_forget(mgainitime)
        log.info(f"Initial time (UTC): {dt} set to receiver")

    # ---------------------------------------------------------------------------------------
    def set_comprot(self, nmea=None, ubx=None):
        comprotcfg = self._poll_port()

        if ubx is None:
            pass
        elif ubx:
            comprotcfg.f.inProtoMask |= 0X01
        else:
            comprotcfg.f.inProtoMask &= ~0X01


        if nmea is None:
            pass
        elif nmea:
            comprotcfg.f.inProtoMask |= 0X02
        else:
            comprotcfg.f.inProtoMask &= ~0X02

        writeResp = self.ubx.set(comprotcfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring communication port: " + str(writeResp))
            raise ConfigurationException()

    def get_comprot(self):
        comprotcfg = self._poll_port()
        ubx = comprotcfg.f.inProtoMask & 0X01 != 0
        nmea = comprotcfg.f.inProtoMask & 0X02 != 0
        return nmea, ubx

    # ---------------------------------------------------------------------------------------
    def get_info(self):
        try:
            rawinfo = self._poll_information()
            model = rawinfo.f.extension_3.split("=")[1]
            fw = rawinfo.f.extension_1.split("=")[1]
            ubxprot = rawinfo.f.extension_2.split("=")[1]
            sat = rawinfo.f.extension_5
            svc = rawinfo.f.extension_6
            sw = rawinfo.f.swVersion
            hw = rawinfo.f.hwVersion
        except:
            log.error("Poll-Error: Cannot get information")
            model = "n/a"
            fw = "n/a"
            ubxprot = "n/a"
            sat = "n/a"
            svc = "n/a"
            sw = "n/a"
            hw = "n/a"
        info = {
            "vendor": "ublox",
            "model": model,
            "fw": fw,
            "ubxprot": ubxprot,
            "sat": sat,
            "svc": svc,
            "sw": sw,
            "hw": hw
        }
        return info

    # ---------------------------------------------------------------------------------------
    def set_alignment(self, yaw, pitch, roll):
        aligncfg = self._poll_alignment()

        if yaw is not None:
            try:
                UbxUtils.range_check(yaw, 0, 360)
                aligncfg.f.yaw = yaw * 100
            except InvalidValue:
                log.error("yaw is not valid, not setting it: %d", yaw)
        if pitch is not None:
            try:
                UbxUtils.range_check(pitch, -90, 90)
                aligncfg.f.pitch = pitch * 100
            except InvalidValue:
                log.error("pitch is not valid, not setting it: %d", pitch)
        if roll is not None:
            try:
                UbxUtils.range_check(roll, -180, 180)
                aligncfg.f.roll = roll * 100
            except InvalidValue:
                log.error("roll is not valid, not setting it: %d", roll)

        writeResp = self.ubx.set(aligncfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring alignment: " + str(writeResp))
            raise ConfigurationException()

    def get_alignment(self):
        cfg = self._poll_alignment()

        # Centidegrees to degrees
        return cfg.f.yaw / 100, cfg.f.pitch / 100, cfg.f.roll / 100


    # ---------------------------------------------------------------------------------------
    def set_positioning(self, vrp2antenna, vrp2imu):
        """ Values in m converted to cm for receiver """
        poscfg = self._poll_positioning()

        if vrp2antenna is not None:
            index = UbxUtils.find_arm_by_name(poscfg, 'vrp2antenna')
            poscfg.f.__setattr__(f'leverArmX_{index}', int(vrp2antenna[0] * 100))
            poscfg.f.__setattr__(f'leverArmY_{index}', int(vrp2antenna[1] * 100))
            poscfg.f.__setattr__(f'leverArmZ_{index}', int(vrp2antenna[2] * 100))

        if vrp2imu is not None:
            index = UbxUtils.find_arm_by_name(poscfg, 'vrp2imu')
            poscfg.f.__setattr__(f'leverArmX_{index}', int(vrp2imu[0] * 100))
            poscfg.f.__setattr__(f'leverArmY_{index}', int(vrp2imu[1] * 100))
            poscfg.f.__setattr__(f'leverArmZ_{index}', int(vrp2imu[2] * 100))

        writeResp = self.ubx.set(poscfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring positioning: " + str(writeResp))
            raise ConfigurationException()

    def get_positioning(self):
        vrp2antenna = []
        vrp2imu = []

        poscfg = self._poll_positioning()

        index = UbxUtils.find_arm_by_name(poscfg, 'vrp2antenna')
        vrp2antenna.append(poscfg.f.get(f'leverArmX_{index}').value / 100.)
        vrp2antenna.append(poscfg.f.get(f'leverArmY_{index}').value / 100.)
        vrp2antenna.append(poscfg.f.get(f'leverArmZ_{index}').value / 100.)

        index = UbxUtils.find_arm_by_name(poscfg, 'vrp2imu')
        vrp2imu.append(poscfg.f.get(f'leverArmX_{index}').value / 100.)
        vrp2imu.append(poscfg.f.get(f'leverArmY_{index}').value / 100.)
        vrp2imu.append(poscfg.f.get(f'leverArmZ_{index}').value / 100.)

        return vrp2antenna, vrp2imu



    # ---------------------------------------------------------------------------------------
    def enable_wheeltick(self):
        self._set_wheeltick_state(True)

    def disable_wheeltick(self):
        self._set_wheeltick_state(False)

    def is_wheeltick_enabled(self):
        gawtcfg = self._poll_gyro_acc_and_wheeltick()
        return gawtcfg.f.flags & UbxCfgEsfGawt.MASK_useWtPin != 0

    def _set_wheeltick_state(self, enable):
        gawtcfg = self._poll_gyro_acc_and_wheeltick()
        if enable:
            gawtcfg.f.flags |= UbxCfgEsfGawt.MASK_useWtPin
        else:
            gawtcfg.f.flags &= ~UbxCfgEsfGawt.MASK_useWtPin

        writeResp = self.ubx.set(gawtcfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring gyro, accelerometer and wheel tick: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def get_nmea_version(self):
        protcfg = self._poll_nav_protocol()
        vProtRead = int(protcfg.f.nmeaVersion)
        vProtReadMain = int(float(vProtRead) / 16)
        vProtReadSub = float(vProtRead - (vProtReadMain * 16))
        vProtReadSub = vProtReadSub / 10
        return float(float(vProtReadMain) + vProtReadSub)

    def set_nmea_version(self, version):
        """ Version is a floating point number """
        protcfg = self._poll_nav_protocol()
        vProt = version
        vProtInt = int(vProt)
        vProtMain = vProtInt * 16
        vProtSub = round((vProt - vProtInt) * 10)
        protcfg.f.nmeaVersion = int(vProtMain + vProtSub)

        writeResp = self.ubx.set(protcfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring navigation protocol: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------
    def set_nav_engine(self, dyn_model=None, fix_mode=None):
        navcfg = self._poll_nav_engine()
        if dyn_model is not None:
            try:
                navcfg.f.dynModel = UbxUtils.nav_engine_model_string2value(dyn_model)
            except InvalidValue:
                log.error("mode value is wrong, not setting it: %s", dyn_model)
        if fix_mode is not None:
            try:
                navcfg.f.fixMode = UbxUtils.get_nav_engine_fix_mode_value(fix_mode)
            except InvalidValue:
                log.error("fix_mode value is wrong, not setting it: %s", fix_mode)

        writeResp = self.ubx.set(navcfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring navigation engine: " + str(writeResp))
            raise ConfigurationException()



    def get_nav_engine(self):
        navcfg = self._poll_nav_engine()
        dyn_model = UbxUtils.nav_engine_model_value2string(navcfg.f.dynModel)
        fix_mode = UbxUtils.get_nav_engine_fix_mode_string(navcfg.f.fixMode)
        return dyn_model, fix_mode

    # ---------------------------------------------------------------------------------------
    def set_navx5(self, adr_enable=None):
        navx5 = self._poll_navx5()
        navx5.f.mask2 = 0
        if adr_enable is not None:
            navx5.f.mask2 |= (1 << 6)
            navx5.f.useAdr = 1 if adr_enable else 0

        writeResp = self.ubx.set(navx5)
        if writeResp is None:
            log.error("nmubxlib response failure when enabling or disabling ADR: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def get_update_rate(self):
        """ Returns update rate in Hz """
        ratecfg = self._poll_update_rate()
        return int(1000 / ratecfg.f.measRate)  # ms to Hz

    def set_update_rate(self, rate):
        """ Update Rate in HZ """
        try:
            UbxUtils.range_check(rate, 1, 2)
        except InvalidValue:
            log.error("update rate value is wrong, not setting it: %d", rate)
            return

        ratecfg = self._poll_update_rate()
        ratecfg.f.measRate = int(1000 / rate)  # Hz to ms
        ratecfg.f.navRate = 1 # cycle. ie synced with measRate. 2 cycle means, every 2 measurement
        writeResp = self.ubx.set(ratecfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring update rate: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def get_high_navigation_rate(self):
        cfg = self._poll_hnr()
        return cfg.get_rate_in_hz()

    def set_high_navigation_rate(self, rate):
        """ High navigation rate """
        try:
            UbxUtils.range_check(rate, 1, 2**8)
        except InvalidValue:
            log.error("high navigation range out of range: %d", rate)

        cfg = self._poll_hnr()
        cfg.set_rate_in_hz(rate)
        writeResp = self.ubx.set(cfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring update rate: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------
    def __switch_gnss_system(self, cfg, value, system):
        if value:
            cfg.enable_gnss(system)
        else:
            cfg.disable_gnss(system)

    def set_gnss_systems(self, gps=None, glonass=None, galileo=None, beidou=None, sbas=None, imes=None, qzss=None):
        cfg = self._poll_gnss_cfg()
        if gps is not None:
            self.__switch_gnss_system(cfg, gps, cfg.GNSS_GPS)
        if galileo is not None:
            self.__switch_gnss_system(cfg, galileo, cfg.GNSS_Galileo)
        if beidou is not None:
            self.__switch_gnss_system(cfg, beidou, cfg.GNSS_BeiDou)
        if glonass is not None:
            self.__switch_gnss_system(cfg, glonass, cfg.GNSS_GLONASS)
        if sbas is not None:
            self.__switch_gnss_system(cfg, sbas, cfg.GNSS_SBAS)
        if imes is not None:
            self.__switch_gnss_system(cfg, imes, cfg.GNSS_IMES)
        if qzss is not None:
            self.__switch_gnss_system(cfg, qzss, cfg.GNSS_QZSS)

        writeResp = self.ubx.set(cfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring gnss systems: " + str(writeResp))
            raise ConfigurationException()

    # ---------------------------------------------------------------------------------------

    def __poll_time_pulse(self):
        poller = UbxCfgTp5Poll()
        # Although NEO-M8L has only one timepulse, we must refer to it as
        # TIMEPULSE2 (id = 1), according to datasheet
        poller.f.tpIdx = 1

        cfg = self.ubx.poll(poller)
        if cfg is None:
            log.error("Polling timepulse config failed")
            raise ConfigurationException()

        return cfg


    def set_timepulse(self, enable=None, frequency=None):
        """ frequency [Hz] """


        cfg = self.__poll_time_pulse()

        if enable is not None:
            if enable:
                cfg.f.flags |= self.TP5_FLAG_ACTIVE
                cfg.f.flags |= self.TP5_FLAG_LOCK_GPS_FREQ
            else:
                cfg.f.flags &= ~self.TP5_FLAG_ACTIVE
                cfg.f.flags &= ~self.TP5_FLAG_LOCK_GPS_FREQ

        if frequency is not None:
            period_us = int(MICROSECONDS_PER_SECOND / frequency)
            cfg.f.flags |= self.TP5_FLAG_IS_LEN
            cfg.f.freqPeriod = period_us
            cfg.f.pulseLenRatio = int(period_us / 2)

        writeResp = self.ubx.set(cfg)
        if writeResp is None:
            log.error("nmubxlib response failure when configuring time pulse")
            raise ConfigurationException()

    def get_timepulse(self):
        cfg = self.__poll_time_pulse()
        enable = True if cfg.f.flags & self.TP5_FLAG_ACTIVE != 0 else False
        if cfg.f.flags & self.TP5_FLAG_IS_LEN != 0:
            freq = int(MICROSECONDS_PER_SECOND / cfg.f.freqPeriod)
        else:
            freq = cfg.f.freqPeriod
        return enable, freq


    # ---------------------------------------------------------------------------------------

    def persist_receiver_config(self):
        ubx_cfg_cfg = UbxCfgCfgAction()
        ubx_cfg_cfg.f.saveMask = ubx_cfg_cfg.MASK_All
        log.info("Persisting receiver configuration")
        self.ubx.set(ubx_cfg_cfg)
